from typing import Optional, List
from .enums import LaserState


class Property:
    def __init__(
        self,
        name: str,
        command: str,
        read_prefix: str = "GET",
        write_prefix: str = "SET",
        write_command: Optional[str] = None,
        read_only: bool = True,
    ):
        self._name = name
        self._command = command
        self._read_only = read_only
        self._read_prefix = read_prefix
        self._write_prefix = write_prefix
        self._write_command = write_command

    def __get__(self, instance, owner):
        msg = instance._query(f"{self._read_prefix}{self._command}")
        return msg

    def __set__(self, instance, value) -> None:
        if self._read_only:
            raise ValueError(f"{self._name} is a read-only attribute")
        else:
            if self._write_command is None:
                instance._write(f"{self._write_prefix}{self._command} {value}")
            else:
                instance._write(f"{self._write_prefix}{self._write_command} {value}")


class FloatProperty(Property):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __get__(self, *args, **kwargs) -> Optional[float]:
        val = super().__get__(*args, **kwargs)
        return float(val)


class IntProperty(Property):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __get__(self, *args, **kwargs) -> Optional[int]:
        val = super().__get__(*args, **kwargs)
        return int(val)


class BoolProperty(Property):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __get__(self, *args, **kwargs) -> Optional[bool]:
        val = super().__get__(*args, **kwargs)
        return bool(int(val))

    def __set__(self, *args, **kwargs):
        if "value" in kwargs:
            kwargs["value"] = int(kwargs["value"])
        else:
            args = list(args)
            args[1] = int(args[1])
        super().__set__(*args, **kwargs)


class FlagProperty(Property):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __get__(self, *args, **kwargs) -> List[bool]:
        ret = super().__get__(*args, **kwargs)
        flags = [bool(int(f)) for f in ret.split(" ")]
        return flags


class LaserStateProperty(IntProperty):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __get__(self, *args, **kwargs) -> LaserState:
        return LaserState(super().__get__(*args, **kwargs))



import logging
from typing import List, Optional

import pyvisa

from .attributes import (
    BoolProperty,
    FlagProperty,
    FloatProperty,
    IntProperty,
    LaserStateProperty,
    Property,
)
from .enums import Alarm, Fault
from .exceptions import MPBCommandError, MPBKeyError


class MPBAmplifier:

    model = Property("Model", "MODEL")
    serial = Property("Serial", "SN")
    enabled = BoolProperty("Laser emission status", "LDenable")
    state = BoolProperty("State", "STATE")
    laser_state = LaserStateProperty("Laser state", "LASERSTATE")
    mode = IntProperty("Mode", "MODE", read_only=False)
    seed_current = FloatProperty("Seed current", "LDCURRENT 1", read_prefix="")
    preamp_current = FloatProperty("Preamp current", "LDCURRENT 2", read_prefix="")
    preamp_current_setpoint = FloatProperty("Preamp current setpoint", "LDCur 2")
    booster_current = FloatProperty("Booster current", "LDCURRENT 3", read_prefix="")
    booster_current_setpoint = FloatProperty(
        "Booster current setpoint",
        "LDCur 3",
        write_prefix="",
        read_only=False,
    )

    shg_temperature = FloatProperty("SHG Temperature", "TECTEMP 4", read_prefix="")
    shg_temperature_setpoint = FloatProperty(
        "SHG Temperature setpoint",
        "TECSETPT 4",
        read_only=False,
    )

    seed_power = FloatProperty("Seed Power", "POWER 3", read_prefix="")
    output_power = FloatProperty("Output Power", "POWER 0", read_prefix="")
    output_power_setpoint = FloatProperty(
        "Output power setpoint", "POWER 0", read_only=False
    )

    power_stabilization = BoolProperty(
        "Power stabilization",
        "POWERENABLE",
        read_prefix="GET",
        write_prefix="",
        read_only=False,
    )

    alarms = FlagProperty("Alarms", "ALR")
    faults = FlagProperty("Faults", "FLT")

    def __init__(self, resource_name: str, baud_rate: int = 9600):
        self.rm = pyvisa.ResourceManager()

        # translate com port to visa port
        if "COM" in resource_name:
            resource_name = f"ASRL{resource_name.strip('COM')}::INSTR"

        self.instr = self.rm.open_resource(
            resource_name,
            baud_rate=baud_rate,
            read_termination="\r",
            write_termination="\r",
        )

    def __exit__(self):
        self.instr.close()
        self.rm.close()

    def _query(self, command: str) -> str:
        msg = self.instr.query(command).strip("D >").strip("F >")
        msg = self._message_error_handling(msg)
        return msg

    def _write(self, command: str) -> None:
        self.instr.write(command)
        msg = self._read()
        self._message_error_handling(msg)

    def _read(self) -> Optional[str]:
        return self.instr.read()

    def _message_error_handling(self, message: str) -> str:
        if "MISSING_ARGUMENT" in message:
            raise MPBCommandError("Missing argument(s)")
        elif "CAN_ONLY_BE_USED_FOR_TESTS" in message:
            raise MPBCommandError("Requires test environment")
        elif "DATA_CANNOT_BE_SET" in message:
            raise MPBCommandError("Cannot execute commanda")
        return message

    def enable_laser(self) -> None:
        try:
            self._write("setLDenable 1")
        except MPBCommandError:
            raise MPBKeyError()

    def disable_laser(self) -> None:
        self._write("setLDenable 0")

    def get_faults(self) -> List[Fault]:
        faults = self.faults
        return [Fault(idx) for idx, flag in enumerate(faults) if flag]

    def get_alarms(self) -> List[Alarm]:
        alarms = self.alarms
        return [Alarm(idx) for idx, flag in enumerate(alarms) if flag]

    def enter_test_environment(self) -> None:
        logging.info("Entering the test environment")
        logging.info(self._query("testeoa"))
        logging.info(self._read())
        logging.info(self._read())
        return

    def save_all(self) -> None:
        """Save settings to non-volatile memory"""
        self._write("SAVEALL")



from enum import IntEnum


class Alarm(IntEnum):
    SHG_TEMPERATURE = 0
    TEC_TEMPERATURE = 1
    PUMP_BIAS = 2
    LOSS_OF_OUTPUT = 3
    CASE_TEMPERATURE = 4


class Fault(IntEnum):
    SHG_TEMPERATURE = 0
    TEC_TEMPERATURE = 1
    LASER_DIODE_CURRENT = 2
    WATCHDOG_TIMEOUT = 3
    CASE_TEMPERATURE = 4


class LaserState(IntEnum):
    OFF = 0
    # The keylock only triggers after the interlock is triggered and the key has to be
    # reset to return the amplifier to normal operation.
    # Not sure how to read out the current key state.
    KEYLOCK = 6
    INTERLOCK = 7
    FAULT = 8
    LOS = 9
    STARTUP = 20
    MANUAL_TURNING_ON = 31
    MANUAL_ON = 41
    SEED_ON = 43
    SEED_OK = 44
    PREAMP_TURN_ON = 45
    PREAMP_TURN_OFF = 46
    PREAMP_ON = 47
    PREAMP_OK = 48
    BOOSTER_TURN_ON = 50
    BOOSTER_TURN_OFF = 51
    BOOSTER_ON = 52
    BOOSTER_OK = 53



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mpbc_vyfa_sf']

package_data = \
{'': ['*']}

install_requires = \
['PyVISA>=1.12.0,<2.0.0']

setup_kwargs = {
    'name': 'mpbc-vyfa-sf',
    'version': '0.2.0',
    'description': 'Python interface for a MPB VYFA amplifier',
    'long_description': '# MPBC-VYFA-SF\n Python interface for a MPB Communications VYFA-SF Series amplifier\n\n# Methods and attributes of the `MPBAmplifier` class\n\n* `enable_laser()`  \n  start laser emission\n* `disable_laser()`  \n  disable laser emission\n* `get_faults()`  \n  get all faults of the amplifier\n* `get_alarms()`  \n  get all alarms of the amplifier\n* `enter_test_environment()`  \n  enter the test environment of the amplifier, required to change the SHG temperature setpoint\n* `save_all()`  \n  save settings to non-volatile memory\n* `model`  \n  amplifier model\n* `serial`  \n  serial number\n* `enabled`  \n  boolean for laser emision status\n* `state`\n* `laser_state`  \n  Current state of the laser; e.g. `BOOSTER_ON`, `OFF` etc. see `enums.py` for more.\n* `mode`\n* `seed_current`  \n  mA\n* `preamp_current`  \n  mA\n* `booster_current`  \n  mA\n* `booster_current_setpoint`  \n  get and set the booster current setpoint\n* `shg_temperature`  \n  C\n* `shg_temperature setpoint`  \n  get and set the shg temperature setpoint in C\n* `seed_power`  \n  mW\n* `output_power`  \n  mW\n* `output_power_setpoint`  \n  if `power_stabilization` is enabled (set to `True`) this gets and sets the output power setpoint in mW\n* `power_stabilization`  \n  enable or disable (`True` or `False`) the output power stabilization. Only settable when emission is disabled.\n\n# Example\n```Python\nfrom mpbc_vyfa_sf import MPBAmplifier\n\namp = MPBAmplifier(com_port = "COM4")\n\n# get the current laser state\namp.laser_state\n\n# get the seed power\namp.seed_power\n\n# enable the laser\namp.enable_laser()\n\n# get the output power\namp.output_power\n\n# disable the laser\namp.disable_laser()\n```\n',
    'author': 'ograsdijk',
    'author_email': 'o.grasdijk@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ograsdijk/MPBC-VYFA-SF',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



class MPBCommandError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class MPBKeyError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__("Put key into enable position.")
